<sdoc-anchor
  id="{{ view_object.render_local_anchor(sdoc_entity) }}"
  node-role="{{ sdoc_entity.get_type_string() }}"
  {% if sdoc_entity.reserved_uid is not none -%}
  data-uid="{{ sdoc_entity.reserved_uid }}"
  {%- endif -%}
>
{%- set incoming_links = view_object.traceability_index.get_incoming_links(sdoc_entity) %}
{%- if incoming_links is not none and incoming_links|length > 0 -%}
<template>
  Incoming link{% if incoming_links|length > 1 -%}s{%- endif %} from:
  {% for incoming_link in incoming_links -%}
  {%- set incoming_link_parent_node = incoming_link.parent_node() -%}
  {%- set incoming_link_href = view_object.render_node_link(incoming_link.parent_node()) -%}
  <a href="{{ incoming_link_href }}">
    {{ incoming_link_parent_node.get_display_title() }}
  </a>
  {% endfor -%}
</template>
{%- endif -%}
</sdoc-anchor>
