{% extends "components/node/index.jinja" %}
{%- set sdoc_entity = node -%}
{%- set section = node -%}
{% block sdoc_entity %}
{% include "components/section/index.jinja" %}
{% endblock sdoc_entity %}
